@@ -287,9 +287,9 @@
                                             (rest values)
                                             values)
                                         table))))
+          (setf (cdr cache) (hash-table-count table))
           (when cl:*compile-verbose*
-            (format t "~&; Float-ops cache prefill: ~D entries~%"
-                    (setf (cdr cache) (hash-table-count table)))))))
+            (format t "~&; Float-ops cache prefill: ~D entries~%" (cdr cache))))))
     table))
 
 (defun record-math-op (key &rest values)
@@ -475,15 +475,19 @@
     (return-from coerce object))
   (with-memoized-math-op (coerce (list object type))
     (if (realp object)
-        (let ((actual-type (if (member type '(double-float long-float))
-                               'double-float
-                               'single-float))
-              (source-value (realnumify object)))
-          (make-flonum (cl:coerce source-value
-                                  (ecase actual-type
-                                    (single-float 'cl:single-float)
-                                    (double-float 'cl:double-float)))
-                       actual-type))
+        (if (and (floatp object)
+                 (eq (flonum-%value object) :minus-zero)
+                 (member type '(double-float single-float)))
+            (make-flonum :minus-zero type)
+            (let ((actual-type (if (member type '(double-float long-float))
+                                   'double-float
+                                   'single-float))
+                  (source-value (realnumify object)))
+              (make-flonum (cl:coerce source-value
+                                      (ecase actual-type
+                                        (single-float 'cl:single-float)
+                                        (double-float 'cl:double-float)))
+                           actual-type)))
         (error "Can't COERCE ~S ~S" object type))))
 
 (macrolet ((define (name)
@@ -774,11 +778,19 @@
 
   ;; Simple case of using the interceptor to return a float.
   ;; As above, no funky values allowed.
-  (intercept (max min + sqrt acos acosh asin asinh atan atanh conjugate cos cosh fceiling ffloor fround ftruncate phase sin sinh tan tanh) (&rest args)
+  (intercept (max min + acos acosh asin asinh atan atanh conjugate cos cosh fceiling ffloor fround ftruncate phase sin sinh tan tanh) (&rest args)
     (dispatch :me
      (make-flonum (apply #':me (realnumify* args))
                   (apply #'pick-result-format args))))
 
+  (intercept (sqrt) (&rest args)
+    (destructuring-bind (x) args
+      (if (zerop x)
+          x
+          (dispatch :me
+                    (make-flonum (funcall #':me (realnumify x))
+                                 (pick-result-format x))))))
+
   (intercept (*) (&rest args)
     (dispatch :me
               (let ((res (make-flonum (apply #':me (realnumify* (substitute-minus-zero args)))
