@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; Data object ref/set stuff.
 
@@ -24,35 +24,38 @@
 
 (define-vop (set-slot)
   (:args (object :scs (descriptor-reg))
-         (value :scs (descriptor-reg any-reg null)))
+         (value :scs (descriptor-reg any-reg zero)))
   (:info name offset lowtag)
   (:ignore name)
   (:results)
   (:generator 1
     (storew value object offset lowtag)))
 
-(define-vop (init-slot set-slot))
-
 (define-vop (compare-and-swap-slot)
   (:args (object :scs (descriptor-reg))
          (old :scs (descriptor-reg any-reg))
          (new :scs (descriptor-reg any-reg)))
   (:info name offset lowtag)
   (:ignore name)
-  (:temporary (:sc interior-reg) lip)
+  (:temporary (:sc non-descriptor-reg) lip)
   (:results (result :scs (descriptor-reg any-reg) :from :load))
   (:generator 5
-    (inst dsb)
     (inst add-sub lip object (- (* offset n-word-bytes) lowtag))
-    LOOP
-    (inst ldxr result lip)
-    (inst cmp result old)
-    (inst b :ne EXIT)
-    (inst stlxr tmp-tn new lip)
-    (inst cbnz tmp-tn LOOP)
-    EXIT
-    (inst clrex)
-    (inst dmb)))
+    (cond ((member :arm-v8.1 *backend-subfeatures*)
+           (move result old)
+           (inst casal result new lip))
+          (t
+           (assemble ()
+             (inst dsb)
+             LOOP
+             (inst ldxr result lip)
+             (inst cmp result old)
+             (inst b :ne EXIT)
+             (inst stlxr tmp-tn new lip)
+             (inst cbnz tmp-tn LOOP)
+             EXIT
+             (inst clrex)
+             (inst dmb))))))
 
 ;;;; Symbol hacking VOPs:
 
@@ -64,74 +67,91 @@
   (:policy :fast-safe)
   (:vop-var vop)
   (:save-p :compute-only))
-;;; Like CHECKED-CELL-REF, only we are a predicate to see if the cell is bound.
-(define-vop (boundp-frob)
-  (:args (object :scs (descriptor-reg)))
-  (:conditional)
-  (:info target not-p)
-  (:policy :fast-safe)
-  (:temporary (:scs (descriptor-reg)) value))
 
-;;; With Symbol-Value, we check that the value isn't the trap object.  So
-;;; Symbol-Value of NIL is NIL.
-#!+sb-thread
+;;; With Symbol-Value, we check that the value isn't the trap object.
+#+sb-thread
 (progn
+  (eval-when (:compile-toplevel)
+    ;; Assert that "CMN reg, 1" is the same as "CMP reg, NO-TLS-VALUE-MARKER"
+    (aver (= (ldb (byte 64 0) -1) no-tls-value-marker)))
+  (defmacro compare-to-no-tls-value-marker (x) `(inst cmn ,x 1))
   (define-vop (set)
-    (:args (object :scs (descriptor-reg))
-           (value :scs (descriptor-reg any-reg null)))
+    (:args (object :scs (descriptor-reg)
+                   :load-if (not (and (sc-is object constant)
+                                      (symbol-always-has-tls-value-p (tn-value object)))))
+           (value :scs (descriptor-reg any-reg zero)))
     (:temporary (:sc any-reg) tls-index)
     (:generator 4
-      (inst ldr (32-bit-reg tls-index) (tls-index-of object))
-      (inst ldr tmp-tn (@ thread-tn tls-index))
-      (inst cmp tmp-tn no-tls-value-marker-widetag)
-      (inst b :ne LOCAL)
-      (storew value object symbol-value-slot other-pointer-lowtag)
-      (inst b DONE)
-      LOCAL
-      (inst str value (@ thread-tn tls-index))
-      DONE))
+      (sc-case object
+        (constant
+         (inst str value (@ thread-tn (make-fixup (tn-value object) :symbol-tls-index))))
+        (t
+         (assemble ()
+           (inst ldr (32-bit-reg tls-index) (tls-index-of object))
+           (inst ldr tmp-tn (@ thread-tn tls-index))
+           (compare-to-no-tls-value-marker tmp-tn)
+           (inst b :ne LOCAL)
+           (storew value object symbol-value-slot other-pointer-lowtag)
+           (inst b DONE)
+           LOCAL
+           (inst str value (@ thread-tn tls-index))
+           DONE)))))
 
   (define-vop (symbol-value checked-cell-ref)
-    (:translate symeval)
+    (:translate symbol-value)
+    (:args (symbol :scs (descriptor-reg) :to :save
+                   :load-if (not (and (sc-is symbol constant)
+                                      (or (symbol-always-has-tls-value-p (tn-value symbol))
+                                          (symbol-always-has-tls-index-p (tn-value symbol)))))))
+    (:args-var symbol-tn-ref)
     (:temporary (:sc any-reg) tls-index)
     (:variant-vars check-boundp)
     (:variant t)
     (:generator 9
-      (inst ldr (32-bit-reg tls-index) (tls-index-of object))
-      (inst ldr value (@ thread-tn tls-index))
-      (inst cmp value no-tls-value-marker-widetag)
-      (inst b :ne LOCAL)
-      (loadw value object symbol-value-slot other-pointer-lowtag)
-      LOCAL
+      (let* ((known-symbol-p (sc-is symbol constant))
+             (known-symbol (and known-symbol-p (tn-value symbol)))
+             (fixup (and known-symbol
+                         (make-fixup known-symbol :symbol-tls-index))))
+        (cond
+          ((symbol-always-has-tls-value-p known-symbol)
+           (inst ldr value (@ thread-tn fixup)))
+          (t
+           (cond
+             (known-symbol              ; e.g. CL:*PRINT-BASE*
+              ;; Known nonzero TLS index, but possibly no per-thread value.
+              ;; The TLS value and global value can be loaded independently.
+              (inst ldr value (@ thread-tn fixup)))
+             (t
+              (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
+              (inst ldr value (@ thread-tn tls-index))))
+
+           (assemble ()
+             (compare-to-no-tls-value-marker value)
+             (inst b :ne LOCAL)
+             (when known-symbol
+               (load-constant vop symbol (setf symbol (tn-ref-load-tn symbol-tn-ref))))
+             (loadw value symbol symbol-value-slot other-pointer-lowtag)
+             LOCAL))))
+
       (when check-boundp
-        (let ((err-lab (generate-error-code vop 'unbound-symbol-error object)))
-          (inst cmp value unbound-marker-widetag)
-          (inst b :eq err-lab)))))
+        (assemble ()
+          (let* ((*location-context* (make-restart-location RETRY value))
+                 (err-lab (generate-error-code vop 'unbound-symbol-error symbol)))
+            (inst cmp value unbound-marker-widetag)
+            (inst b :eq err-lab))
+          RETRY))))
 
   (define-vop (fast-symbol-value symbol-value)
     (:policy :fast)
     (:variant nil)
-    (:variant-cost 5))
+    (:variant-cost 5)))
 
-  (define-vop (boundp boundp-frob)
-    (:translate boundp)
-    (:temporary (:sc any-reg) tls-index)
-    (:generator 9
-      (inst ldr (32-bit-reg tls-index) (tls-index-of object))
-      (inst ldr value (@ thread-tn tls-index))
-      (inst cmp value no-tls-value-marker-widetag)
-      (inst b :ne LOCAL)
-      (loadw value object symbol-value-slot other-pointer-lowtag)
-      LOCAL
-      (inst cmp value unbound-marker-widetag)
-      (inst b (if not-p :eq :ne) target))))
-
-#!-sb-thread
+#-sb-thread
 (progn
   (define-vop (set cell-set)
     (:variant symbol-value-slot other-pointer-lowtag))
   (define-vop (symbol-value checked-cell-ref)
-    (:translate symeval)
+    (:translate symbol-value)
     (:generator 9
                 (loadw value object symbol-value-slot other-pointer-lowtag)
                 (let ((err-lab (generate-error-code vop 'unbound-symbol-error object)))
@@ -140,14 +160,30 @@
   (define-vop (fast-symbol-value cell-ref)
     (:variant symbol-value-slot other-pointer-lowtag)
     (:policy :fast)
-    (:translate symeval))
+    (:translate symbol-value)))
 
-  (define-vop (boundp boundp-frob)
-    (:translate boundp)
-    (:generator 9
+(define-vop (boundp)
+  (:args (object :scs (descriptor-reg)))
+  (:conditional)
+  (:info target not-p)
+  (:policy :fast-safe)
+  (:temporary (:scs (descriptor-reg)) value)
+  (:translate boundp)
+  #+sb-thread
+  (:generator 9
+      (inst ldr (32-bit-reg value) (tls-index-of object))
+      (inst ldr value (@ thread-tn value))
+      (compare-to-no-tls-value-marker value)
+      (inst b :ne LOCAL)
+      (loadw value object symbol-value-slot other-pointer-lowtag)
+      LOCAL
+      (inst cmp value unbound-marker-widetag)
+      (inst b (if not-p :eq :ne) target))
+  #-sb-thread
+  (:generator 9
       (loadw value object symbol-value-slot other-pointer-lowtag)
       (inst cmp value unbound-marker-widetag)
-      (inst b (if not-p :eq :ne) target))))
+      (inst b (if not-p :eq :ne) target)))
 
 (define-vop (%set-symbol-global-value cell-set)
   (:variant symbol-value-slot other-pointer-lowtag))
@@ -155,11 +191,11 @@
 (define-vop (fast-symbol-global-value cell-ref)
   (:variant symbol-value-slot other-pointer-lowtag)
   (:policy :fast)
-  (:translate sym-global-val))
+  (:translate symbol-global-value))
 
 (define-vop (symbol-global-value)
   (:policy :fast-safe)
-  (:translate sym-global-val)
+  (:translate symbol-global-value)
   (:args (object :scs (descriptor-reg) :to (:result 1)))
   (:results (value :scs (descriptor-reg any-reg)))
   (:vop-var vop)
@@ -174,16 +210,38 @@
   (:policy :fast-safe)
   (:translate symbol-hash)
   (:args (symbol :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) temp)
+  (:args-var args)
   (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
   (:generator 2
     ;; The symbol-hash slot of NIL holds NIL because it is also the
-    ;; cdr slot, so we have to strip off the two low bits to make sure
+    ;; car slot, so we have to strip off the fixnum-tag-mask to make sure
     ;; it is a fixnum.  The lowtag selection magic that is required to
     ;; ensure this is explained in the comment in objdef.lisp
-    (loadw temp symbol symbol-hash-slot other-pointer-lowtag)
-    (inst and res temp (bic-mask fixnum-tag-mask))))
+    (loadw res symbol symbol-hash-slot other-pointer-lowtag)
+    (unless (not-nil-tn-ref-p args)
+      (inst and res res (bic-mask fixnum-tag-mask)))))
+
+(define-vop ()
+  (:args (symbol :scs (descriptor-reg)))
+  (:results (result :scs (unsigned-reg)))
+  (:result-types positive-fixnum)
+  (:translate symbol-package-id)
+  (:policy :fast-safe)
+  (:generator 1
+    #.(assert (= sb-impl::package-id-bits 16))
+    (inst ldrh result (@ symbol (+ (ash symbol-name-slot word-shift)
+                                  (- other-pointer-lowtag)
+                                  6))))) ; little-endian
+(define-vop ()
+  (:policy :fast-safe)
+  (:translate symbol-name)
+  (:args (symbol :scs (descriptor-reg)))
+  (:results (result :scs (descriptor-reg)))
+  (:generator 5
+    (pseudo-atomic (tmp-tn :sync nil)
+      (loadw tmp-tn symbol symbol-name-slot other-pointer-lowtag)
+      (inst and result tmp-tn (1- (ash 1 sb-impl::symbol-name-bits))))))
 
 (define-vop (%compare-and-swap-symbol-value)
   (:translate %compare-and-swap-symbol-value)
@@ -191,14 +249,14 @@
          (old :scs (descriptor-reg any-reg))
          (new :scs (descriptor-reg any-reg)))
   (:results (result :scs (descriptor-reg any-reg) :from :load))
-  #!+sb-thread
+  #+sb-thread
   (:temporary (:sc any-reg) tls-index)
-  (:temporary (:sc interior-reg) lip)
+  (:temporary (:sc non-descriptor-reg) lip)
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 15
     (inst dsb)
-    #!+sb-thread
+    #+sb-thread
     (assemble ()
       (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
       ;; Thread-local area, no synchronization needed.
@@ -208,7 +266,7 @@
       (inst str new (@ thread-tn tls-index))
       DONT-STORE-TLS
 
-      (inst cmp result no-tls-value-marker-widetag)
+      (compare-to-no-tls-value-marker result)
       (inst b :ne CHECK-UNBOUND))
     (inst add-sub lip symbol (- (* symbol-value-slot n-word-bytes)
                                 other-pointer-lowtag))
@@ -226,6 +284,40 @@
     (inst dmb)
     (inst cmp result unbound-marker-widetag)
     (inst b :eq (generate-error-code vop 'unbound-symbol-error symbol))))
+
+(define-vop (%compare-and-swap-symbol-value-v8.1)
+  (:translate %compare-and-swap-symbol-value)
+  (:args (symbol :scs (descriptor-reg))
+         (old :scs (descriptor-reg any-reg))
+         (new :scs (descriptor-reg any-reg)))
+  (:results (result :scs (descriptor-reg any-reg) :from :load))
+  #+sb-thread
+  (:temporary (:sc any-reg) tls-index)
+  (:temporary (:sc non-descriptor-reg) lip)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:guard (member :arm-v8.1 *backend-subfeatures*))
+  (:generator 14
+    #+sb-thread
+    (assemble ()
+      (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
+      ;; Thread-local area, no synchronization needed.
+      (inst ldr result (@ thread-tn tls-index))
+      (inst cmp result old)
+      (inst b :ne DONT-STORE-TLS)
+      (inst str new (@ thread-tn tls-index))
+      DONT-STORE-TLS
+
+      (compare-to-no-tls-value-marker result)
+      (inst b :ne CHECK-UNBOUND))
+    (inst add-sub lip symbol (- (* symbol-value-slot n-word-bytes)
+                                other-pointer-lowtag))
+    (move result old)
+    (inst casal result new lip)
+    CHECK-UNBOUND
+    (inst cmp result unbound-marker-widetag)
+    (inst b :eq (generate-error-code vop 'unbound-symbol-error symbol))))
+
 
 ;;;; Fdefinition (fdefn) objects.
 
@@ -252,16 +344,16 @@
   (:translate (setf fdefn-fun))
   (:args (function :scs (descriptor-reg) :target result)
          (fdefn :scs (descriptor-reg)))
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:scs (non-descriptor-reg)) lip)
   (:temporary (:scs (non-descriptor-reg)) type)
   (:results (result :scs (descriptor-reg)))
   (:generator 38
-    (inst add-sub lip function (- (* simple-fun-code-offset n-word-bytes)
+    (inst add-sub lip function (- (* simple-fun-insts-offset n-word-bytes)
                                   fun-pointer-lowtag))
     (load-type type function (- fun-pointer-lowtag))
     (inst cmp type simple-fun-widetag)
     (inst b :eq SIMPLE-FUN)
-    (load-inline-constant lip '(:fixup closure-tramp :assembly-routine) lip)
+    (load-inline-constant lip '(:fixup closure-tramp :assembly-routine))
     SIMPLE-FUN
     (storew lip fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
@@ -270,16 +362,12 @@
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
-  (:args (fdefn :scs (descriptor-reg) :target result))
+  (:args (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) temp)
-  (:temporary (:scs (interior-reg)) lip)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
     (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
-    (load-inline-constant temp '(:fixup undefined-tramp :assembly-routine) lip)
-    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-    (move result fdefn)))
-
+    (load-inline-constant temp '(:fixup undefined-tramp :assembly-routine))
+    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 
 ;;;; Binding and Unbinding.
@@ -287,41 +375,54 @@
 ;;; BIND -- Establish VAL as a binding for SYMBOL.  Save the old value and
 ;;; the symbol on the binding stack and stuff the new value into the
 ;;; symbol.
-#!+sb-thread
+#+sb-thread
 (progn
   (define-vop (dynbind)
-    (:args (value :scs (any-reg descriptor-reg) :to :save)
-           (symbol :scs (descriptor-reg)))
-    (:temporary (:sc descriptor-reg) value-temp)
-    (:temporary (:sc descriptor-reg :offset r0-offset :from (:argument 1)) alloc-tls-symbol)
+    (:args (value :scs (any-reg descriptor-reg zero) :to :save)
+           (symbol :scs (descriptor-reg)
+                   :load-if (not (and (sc-is symbol constant)
+                                      (or (symbol-always-has-tls-value-p (tn-value symbol))
+                                          (symbol-always-has-tls-index-p (tn-value symbol)))))
+                   :target alloc-tls-symbol))
+    (:temporary (:sc descriptor-reg :offset r8-offset :from (:argument 1)) alloc-tls-symbol)
     (:temporary (:sc non-descriptor-reg :offset nl0-offset) tls-index)
     (:temporary (:sc non-descriptor-reg :offset nl1-offset) free-tls-index)
-    (:temporary (:sc interior-reg) lip)
+    (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
     (:ignore free-tls-index)
     (:temporary (:scs (any-reg)) bsp)
      (:generator 5
       (load-binding-stack-pointer bsp)
-      (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
       (inst add bsp bsp (* binding-size n-word-bytes))
       (store-binding-stack-pointer bsp)
-      (inst cbnz (32-bit-reg tls-index) TLS-INDEX-VALID)
-      (move alloc-tls-symbol symbol)
-      (load-inline-constant value-temp '(:fixup alloc-tls-index :assembly-routine) lip)
-      (inst blr value-temp)
-      TLS-INDEX-VALID
+      (let* ((known-symbol-p (sc-is symbol constant))
+             (known-symbol (and known-symbol-p (tn-value symbol)))
+             (tls-index-reg tls-index)
+             (tls-index (if known-symbol
+                            (make-fixup known-symbol :symbol-tls-index)
+                            tls-index)))
+        (assemble ()
+          (cond (known-symbol
+                 (inst movz tls-index-reg tls-index))
+                (t
+                 (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
+                 (inst cbnz (32-bit-reg tls-index) TLS-INDEX-VALID)
+                 (move alloc-tls-symbol symbol)
+                 (load-inline-constant lr '(:fixup alloc-tls-index :assembly-routine))
+                 (inst blr lr)))
+          TLS-INDEX-VALID
+          (inst ldr alloc-tls-symbol (@ thread-tn tls-index))
+          (inst stp alloc-tls-symbol tls-index-reg
+                (@ bsp (* (- binding-value-slot binding-size)
+                          n-word-bytes)))
+          (inst str value (@ thread-tn tls-index))))))
 
-      (inst ldr value-temp (@ thread-tn tls-index))
-      (inst stp value-temp tls-index (@ bsp (* (- binding-value-slot binding-size)
-                                               n-word-bytes)))
-      (inst str value (@ thread-tn tls-index))))
-
-  (define-vop (unbind)
-    (:info n)
+  (define-vop (unbind-n)
+    (:info symbols)
     (:temporary (:sc descriptor-reg) value)
     (:temporary (:sc any-reg) tls-index bsp)
     (:generator 0
       (load-binding-stack-pointer bsp)
-      (loop repeat n
+      (loop for symbol in symbols
             do
             (inst ldp value tls-index (@ bsp (* (- binding-value-slot binding-size)
                                                 n-word-bytes)))
@@ -334,14 +435,9 @@
                                          :pre-index)))
       (store-binding-stack-pointer bsp))))
 
-(define-vop (unbind-to-here)
-  (:args (arg :scs (descriptor-reg any-reg) :target where))
-  (:temporary (:scs (any-reg) :from (:argument 0)) where)
-  (:temporary (:scs (descriptor-reg)) symbol value)
-  (:temporary (:scs (any-reg)) bsp)
-  (:generator 0
+(defun unbind-to-here (where symbol value bsp)
+  (assemble ()
     (load-binding-stack-pointer bsp)
-    (move where arg)
     (inst cmp where bsp)
     (inst b :eq DONE)
 
@@ -350,9 +446,9 @@
     (inst ldp value symbol (@ bsp (* (- binding-value-slot binding-size)
                                      n-word-bytes)))
     (inst cbz symbol ZERO)
-    #!-sb-thread
+    #-sb-thread
     (storew value symbol symbol-value-slot other-pointer-lowtag)
-    #!+sb-thread
+    #+sb-thread
     (inst str value (@ thread-tn symbol))
     ZERO
     (inst stp zr-tn zr-tn (@ bsp (* (- binding-value-slot binding-size)
@@ -364,7 +460,17 @@
 
     DONE
     (store-binding-stack-pointer bsp)))
-#!-sb-thread
+
+(define-vop (unbind-to-here)
+  (:args (arg :scs (descriptor-reg any-reg) :target where))
+  (:temporary (:scs (any-reg) :from (:argument 0)) where)
+  (:temporary (:scs (descriptor-reg)) symbol value)
+  (:temporary (:scs (any-reg)) bsp)
+  (:generator 0
+    (move where arg)
+    (unbind-to-here where symbol value bsp)))
+
+#-sb-thread
 (progn
   (define-vop (dynbind)
     (:args (val :scs (any-reg descriptor-reg))
@@ -400,19 +506,27 @@
   closure-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %closure-index-ref)
 
-(define-full-setter set-funcallable-instance-info *
-  funcallable-instance-info-offset fun-pointer-lowtag
-  (descriptor-reg any-reg null) * %set-funcallable-instance-info)
+(define-full-setter %closure-index-set *
+  closure-info-offset fun-pointer-lowtag
+  (descriptor-reg any-reg) * %closure-index-set)
 
 (define-full-reffer funcallable-instance-info *
   funcallable-instance-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %funcallable-instance-info)
 
-(define-vop (closure-ref slot-ref)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-ref)
+  (:args (object :scs (descriptor-reg)))
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (loadw value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
-(define-vop (closure-init slot-set)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-init)
+  (:args (object :scs (descriptor-reg))
+         (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (storew value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
 (define-vop (closure-init-from-fp)
   (:args (object :scs (descriptor-reg)))
@@ -430,36 +544,86 @@
 
 ;;;; Instance hackery:
 
-(define-vop (instance-length)
+(define-vop ()
   (:policy :fast-safe)
   (:translate %instance-length)
   (:args (struct :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) temp)
   (:results (res :scs (unsigned-reg)))
   (:result-types positive-fixnum)
   (:generator 4
-    (loadw temp struct 0 instance-pointer-lowtag)
-    (inst lsr res temp n-widetag-bits)))
+    (loadw res struct 0 instance-pointer-lowtag)
+    (inst lsr res res instance-length-shift)))
 
 (define-full-reffer instance-index-ref * instance-slots-offset
   instance-pointer-lowtag (descriptor-reg any-reg) * %instance-ref)
 
 (define-full-setter instance-index-set * instance-slots-offset
-  instance-pointer-lowtag (descriptor-reg any-reg null) * %instance-set)
+  instance-pointer-lowtag (descriptor-reg any-reg) * %instance-set)
 
 (define-vop (%instance-cas word-index-cas)
   (:policy :fast-safe)
   (:translate %instance-cas)
   (:variant instance-slots-offset instance-pointer-lowtag)
   (:arg-types instance tagged-num * *))
+(define-vop (%raw-instance-cas/word %instance-cas)
+  (:args (object)
+         (index)
+         (old-value :scs (unsigned-reg))
+         (new-value :scs (unsigned-reg)))
+  (:arg-types * tagged-num unsigned-num unsigned-num)
+  (:results (result :scs (unsigned-reg) :from :load))
+  (:result-types unsigned-num)
+  (:translate %raw-instance-cas/word))
+
+(define-vop (%raw-instance-cas/signed-word %instance-cas)
+  (:args (object)
+         (index)
+         (old-value :scs (signed-reg))
+         (new-value :scs (signed-reg)))
+  (:arg-types * tagged-num signed-num signed-num)
+  (:results (result :scs (signed-reg) :from :load))
+  (:result-types signed-num)
+  (:translate %raw-instance-cas/signed-word))
+
 
 ;;;; Code object frobbing.
 
 (define-full-reffer code-header-ref * 0 other-pointer-lowtag
   (descriptor-reg any-reg) * code-header-ref)
 
-(define-full-setter code-header-set * 0 other-pointer-lowtag
-  (descriptor-reg any-reg null) * code-header-set)
+#-darwin-jit
+(define-vop (code-header-set)
+  (:translate code-header-set)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg))
+         (value :scs (any-reg descriptor-reg)))
+  (:arg-types * tagged-num *)
+  (:temporary (:scs (non-descriptor-reg)) temp card)
+  (:temporary (:sc non-descriptor-reg) pa-flag)
+  (:generator 10
+    (load-inline-constant temp `(:fixup "gc_card_table_mask" :foreign-dataref))
+    (inst ldr temp (@ temp))
+    (inst ldr (32-bit-reg temp) (@ temp)) ; 4-byte int
+    (pseudo-atomic (pa-flag)
+      ;; Compute card mark index
+      (inst lsr card object gencgc-card-shift)
+      (inst and card card temp)
+      ;; Load mark table base
+      (load-inline-constant temp `(:fixup "gc_card_mark" :foreign-dataref))
+      (inst ldr temp (@ temp))
+      (inst ldr temp (@ temp))
+      ;; Touch the card mark byte.
+      (inst strb null-tn (@ temp card))
+      ;; set 'written' flag in the code header
+      ;; If two threads get here at the same time, they'll write the same byte.
+      (let ((byte (- #+big-endian 4 #+little-endian 3 other-pointer-lowtag)))
+        (inst ldrb temp (@ object byte))
+        (inst orr temp temp #x40)
+        (inst strb temp (@ object byte)))
+      (inst lsl temp index (- word-shift n-fixnum-tag-bits))
+      (inst sub temp temp other-pointer-lowtag)
+      (inst str value (@ object temp)))))
 
 ;;;; raw instance slot accessors
 
@@ -483,11 +647,9 @@
                        (inst ,instruction value (@ object offset))))
                     ,@(when move-result
                         `((,move-macro result value))))))
-         (let ((ref-vop (symbolicate "RAW-INSTANCE-REF/" name))
-               (set-vop (symbolicate "RAW-INSTANCE-SET/" name)))
            `(progn
-              (define-vop (,ref-vop)
-                (:translate ,(symbolicate "%" ref-vop))
+              (define-vop ()
+                (:translate ,(symbolicate "%RAW-INSTANCE-REF/" name))
                 (:policy :fast-safe)
                 (:args (object :scs (descriptor-reg))
                        (index :scs (any-reg immediate)))
@@ -496,17 +658,15 @@
                 (:result-types ,value-primtype)
                 (:temporary (:scs (non-descriptor-reg)) offset)
                 (:generator 5 ,@(emit-generator 'ldr nil)))
-              (define-vop (,set-vop)
-                (:translate ,(symbolicate "%" set-vop))
+              (define-vop ()
+                (:translate ,(symbolicate "%RAW-INSTANCE-SET/" name))
                 (:policy :fast-safe)
                 (:args (object :scs (descriptor-reg))
                        (index :scs (any-reg immediate))
-                       (value :scs (,value-sc) :target result))
+                       (value :scs (,value-sc)))
                 (:arg-types * positive-fixnum ,value-primtype)
-                (:results (result :scs (,value-sc)))
-                (:result-types ,value-primtype)
                 (:temporary (:scs (non-descriptor-reg)) offset)
-                (:generator 5 ,@(emit-generator 'str t))))))))
+                (:generator 5 ,@(emit-generator 'str nil)))))))
   (define-raw-slot-vops word unsigned-num unsigned-reg)
   (define-raw-slot-vops signed-word signed-num signed-reg)
   (define-raw-slot-vops single single-float single-reg
@@ -526,7 +686,7 @@
          (diff :scs (unsigned-reg)))
   (:arg-types * positive-fixnum unsigned-num)
   (:temporary (:sc non-descriptor-reg) sum)
-  (:temporary (:sc interior-reg) lip)
+  (:temporary (:sc non-descriptor-reg) lip)
   (:results (result :scs (unsigned-reg) :from :load))
   (:result-types unsigned-num)
   (:generator 4
@@ -542,3 +702,22 @@
     (inst stlxr tmp-tn sum lip)
     (inst cbnz tmp-tn LOOP)
     (inst dmb)))
+
+(define-vop (raw-instance-atomic-incf/word-v8.1)
+  (:translate %raw-instance-atomic-incf/word)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg))
+         (diff :scs (unsigned-reg)))
+  (:arg-types * positive-fixnum unsigned-num)
+  (:temporary (:sc non-descriptor-reg) lip)
+  (:results (result :scs (unsigned-reg) :from :load))
+  (:result-types unsigned-num)
+  (:guard (member :arm-v8.1 *backend-subfeatures*))
+  (:generator 3
+    (inst add lip object (lsl index (- word-shift n-fixnum-tag-bits)))
+    (inst add lip lip (- (* instance-slots-offset
+                            n-word-bytes)
+                         instance-pointer-lowtag))
+
+    (inst ldaddal diff result lip)))
