@@ -322,7 +322,7 @@
          #+sb-thread
          (progn
            (when ,sync
-            (inst dmb))
+            (inst dmb :ishst))
            (inst str (32-bit-reg zr-tn)
                  (@ thread-tn
                     (* n-word-bytes thread-pseudo-atomic-bits-slot)))
@@ -345,7 +345,8 @@
      (:args (object :scs (descriptor-reg))
             (index :scs (any-reg immediate)))
      (:arg-types ,type tagged-num)
-     (:temporary (:scs (non-descriptor-reg)) lip)
+     (:temporary (:scs (non-descriptor-reg)
+                  :unused-if (sc-is index immediate)) lip)
      (:results (value :scs ,scs))
      (:result-types ,el-type)
      (:generator 5
@@ -368,7 +369,8 @@
             (index :scs (any-reg immediate))
             (value :scs (,@scs zero)))
      (:arg-types ,type tagged-num ,el-type)
-     (:temporary (:scs (non-descriptor-reg)) lip)
+     (:temporary (:scs (non-descriptor-reg)
+                  :unused-if (sc-is index immediate)) lip)
      (:generator 2
        (sc-case index
          (immediate
@@ -390,7 +392,8 @@
      (:arg-types ,type tagged-num)
      (:results (value :scs ,scs))
      (:result-types ,el-type)
-     (:temporary (:scs (non-descriptor-reg)) lip)
+     (:temporary (:scs (non-descriptor-reg)
+                  :unused-if (sc-is index immediate)) lip)
      (:generator 5
        ,@(multiple-value-bind (op shift)
              (ecase size
@@ -399,7 +402,9 @@
                (:short
                 (values (if signed 'ldrsh 'ldrh) 1))
                (:word
-                (values (if signed 'ldrsw 'ldr) 2)))
+                (values (if signed 'ldrsw 'ldr) 2))
+               (:single-float
+                (values 'ldr 2)))
            (let ((value (if (and (eq size :word)
                                  (not signed))
                             '(32-bit-reg value)
@@ -423,53 +428,61 @@
 
 (defmacro define-partial-setter (name type size offset lowtag scs el-type
                                  &optional translate)
-  `(define-vop (,name)
-     ,@(when translate
-             `((:translate ,translate)))
-     (:policy :fast-safe)
-     (:args (object :scs (descriptor-reg))
-            (index :scs (any-reg unsigned-reg immediate))
-            (value :scs ,scs
-                   :load-if (not (and (sc-is value immediate)
-                                      (eql (tn-value value) 0)))))
-     (:arg-types ,type tagged-num ,el-type)
-     (:temporary (:scs (non-descriptor-reg)) lip)
-     (:generator 5
-       (when (sc-is value immediate)
-         (setf value zr-tn))
-       ,@(multiple-value-bind (op shift)
-             (ecase size
-               (:byte
-                (values 'strb 0))
-               (:short
-                (values 'strh 1))
-               (:word
-                (values 'str 2)))
-           (let ((value (if (eq size :word)
-                            '(32-bit-reg value)
-                            'value)))
-             `((sc-case index
-                 (immediate
-                  (inst ,op ,value (@ object (load-store-offset
-                                              (+
-                                               (ash (tn-value index) ,shift)
-                                               (- (* ,offset n-word-bytes) ,lowtag))))))
-                 (t
-                  (let ((shift ,shift))
-                    (sc-case index
-                      (any-reg
-                       (decf shift n-fixnum-tag-bits)))
-                    (inst add lip object (if (minusp shift)
-                                             (asr index (- shift))
-                                             (lsl index shift)))
-                    (inst ,op
-                          ,value (@ lip (- (* ,offset n-word-bytes) ,lowtag))))))))))))
+  (let ((value `((value :scs ,scs
+                             :load-if (not (and (sc-is value immediate)
+                                                (eql (tn-value value) 0))))))
+        (setf-p (typep translate '(cons (eql setf)))))
+   `(define-vop (,name)
+      ,@(when translate
+          `((:translate ,translate)))
+      (:policy :fast-safe)
+      (:args ,@(when setf-p
+                 value)
+             (object :scs (descriptor-reg))
+             (index :scs (any-reg unsigned-reg immediate))
+             ,@(unless setf-p
+                 value))
+      (:arg-types ,@(when setf-p
+                      `(,el-type))
+                  ,type
+                  tagged-num
+                  ,@(unless setf-p
+                      `(,el-type)))
+      (:temporary (:scs (non-descriptor-reg)
+                   :unused-if (sc-is index immediate)) lip)
+      (:generator 5
+        (when (sc-is value immediate)
+          (setf value zr-tn))
+        ,@(multiple-value-bind (op shift)
+              (ecase size
+                (:byte
+                 (values 'strb 0))
+                (:short
+                 (values 'strh 1))
+                ((:word :single-float)
+                 (values 'str 2)))
+            (let ((value (if (eq size :word)
+                             '(32-bit-reg value)
+                             'value)))
+              `((sc-case index
+                  (immediate
+                   (inst ,op ,value (@ object (load-store-offset
+                                               (+
+                                                (ash (tn-value index) ,shift)
+                                                (- (* ,offset n-word-bytes) ,lowtag))))))
+                  (t
+                   (let ((shift ,shift))
+                     (sc-case index
+                       (any-reg
+                        (decf shift n-fixnum-tag-bits)))
+                     (inst add lip object (if (minusp shift)
+                                              (asr index (- shift))
+                                              (lsl index shift)))
+                     (inst ,op
+                           ,value (@ lip (- (* ,offset n-word-bytes) ,lowtag)))))))))))))
 
-(defun load-inline-constant (dst value)
-  (destructuring-bind (size . label) (register-inline-constant value)
-    (ecase size
-      (:qword
-       (inst load-from-label dst label)))))
+(defun load-inline-constant (dst &rest constant-descriptor)
+  (inst load-from-label dst (cdr (apply #'register-inline-constant constant-descriptor))))
 
 ;;;
 
