@@ -15,20 +15,17 @@
 
 ;;;; LAMBDA hackery
 
-;;;; FIXME: where is that file?
-;;;; Answer: As usual, with CMUCL.
 ;;;; Note: Take a look at the compiler-overview.tex section on "Hairy
 ;;;; function representation" before you seriously mess with this
 ;;;; stuff.
-#-sb-devel
 (declaim (start-block ir1-convert-lambda ir1-convert-lambda-body
                       ir1-convert-aux-bindings varify-lambda-arg
                       ir1-convert-lambdalike))
 
 ;;; Return a VAR structure for NAME, filling in info if it is globally
 ;;; special. If it is losing, we punt with a COMPILER-ERROR.
-(declaim (ftype (sfunction (symbol) lambda-var) varify-lambda-arg))
-(defun varify-lambda-arg (name)
+(declaim (ftype (sfunction (symbol &optional t) lambda-var) varify-lambda-arg))
+(defun varify-lambda-arg (name &optional source-form)
   (case (info :variable :kind name)
     (:special
      (let ((variable (find-free-var name)))
@@ -37,7 +34,8 @@
                         :where-from (leaf-where-from variable)
                         :specvar variable)))
     (t
-     (make-lambda-var :%source-name name))))
+     (make-lambda-var :%source-name name
+                      :source-form source-form))))
 
 ;;; Parse a lambda list into a list of VAR structures, stripping off
 ;;; any &AUX bindings. Each arg name is checked for legality, and
@@ -118,7 +116,7 @@
 ;;; SOURCE-NAME and DEBUG-NAME. But I (WHN) don't use &AUX bindings,
 ;;; so I'm not motivated. Patches will be accepted...
 (defun ir1-convert-aux-bindings (start next result body aux-vars aux-vals
-                                 post-binding-lexenv)
+                                 post-binding-lexenv &key value-source-forms)
   (declare (type ctran start next) (type (or lvar null) result)
            (list body aux-vars aux-vals))
   (if (null aux-vars)
@@ -133,10 +131,12 @@
                                           :post-binding-lexenv post-binding-lexenv
                                           :debug-name (debug-name
                                                        '&aux-bindings
-                                                       aux-vars))))
+                                                       aux-vars)
+                                          :value-source-forms (rest value-source-forms))))
         (reference-leaf start ctran fun-lvar fun)
         (ir1-convert-combination-args fun-lvar ctran next result
-                                      (list (first aux-vals)))))
+                                      (list (first aux-vals))
+                                      :arg-source-forms (list (first value-source-forms)))))
   (values))
 
 ;;; This is similar to IR1-CONVERT-PROGN-BODY except that code to bind
@@ -151,13 +151,15 @@
 ;;; to start a block outside of this cleanup, causing cleanup code to
 ;;; be emitted when the scope is exited.
 (defun ir1-convert-special-bindings
-    (start next result body aux-vars aux-vals svars post-binding-lexenv)
+    (start next result body aux-vars aux-vals svars post-binding-lexenv
+     &key value-source-forms)
   (declare (type ctran start next) (type (or lvar null) result)
            (list body aux-vars aux-vals svars))
   (cond
    ((null svars)
     (ir1-convert-aux-bindings start next result body aux-vars aux-vals
-                              post-binding-lexenv))
+                              post-binding-lexenv
+                              :value-source-forms value-source-forms))
    (t
     (ctran-starts-block next)
     (let ((cleanup (make-cleanup :kind :special-bind))
@@ -165,14 +167,15 @@
           (bind-ctran (make-ctran))
           (cleanup-ctran (make-ctran)))
       (ir1-convert start bind-ctran nil
-                   `(%special-bind ',(leaf-source-name (lambda-var-specvar var)) ,var))
+                   `(%special-bind ',(lambda-var-specvar var) ,var))
       (setf (cleanup-mess-up cleanup) (ctran-use bind-ctran))
       (let ((*lexenv* (make-lexenv :cleanup cleanup)))
         (ir1-convert bind-ctran cleanup-ctran nil '(%cleanup-point))
         (ir1-convert-special-bindings cleanup-ctran next result
                                       body aux-vars aux-vals
                                       (rest svars)
-                                      post-binding-lexenv)))))
+                                      post-binding-lexenv
+                                      :value-source-forms value-source-forms)))))
   (values))
 
 ;;; Create a lambda node out of some code, returning the result. The
@@ -201,7 +204,9 @@
                                 debug-name
                                 (note-lexical-bindings t)
                                 post-binding-lexenv
-                                system-lambda)
+                                system-lambda
+                                local-policy
+                                value-source-forms)
   (declare (list body vars aux-vars aux-vals))
 
   ;; We're about to try to put new blocks into *CURRENT-COMPONENT*.
@@ -209,10 +214,13 @@
 
   (let* ((bind (make-bind))
          (lambda (make-lambda :vars vars
-                  :bind bind
-                  :%source-name source-name
-                  :%debug-name debug-name
-                  :system-lambda-p system-lambda))
+                              :bind bind
+                              :%source-name source-name
+                              :%debug-name debug-name
+                              :system-lambda-p system-lambda
+                              :lexenv (if local-policy
+                                          (make-lexenv :policy local-policy)
+                                          *lexenv*)))
          (result-ctran (make-ctran))
          (result-lvar (make-lvar)))
     ;; just to check: This function should fail internal assertions if
@@ -267,7 +275,8 @@
             (ir1-convert-special-bindings postbind-ctran result-ctran
                                           result-lvar body
                                           aux-vars aux-vals (svars)
-                                          post-binding-lexenv)))))
+                                          post-binding-lexenv
+                                          :value-source-forms value-source-forms)))))
 
     (link-blocks (component-head *current-component*) (node-block bind))
     (push lambda (component-new-functionals *current-component*))
@@ -307,7 +316,7 @@
                 (dolist (default defaults)
                   (if (constantp default)
                       (default-vals default)
-                      (let ((var (sb-xc:gensym)))
+                      (let ((var (gensym)))
                         (default-bindings `(,var ,default))
                         (default-vals var))))
                 (let ((bindings (default-bindings))
@@ -447,10 +456,10 @@
                                  :type (leaf-type var)
                                  :where-from (leaf-where-from var))))
 
-    (let* ((n-context (sb-xc:gensym "N-CONTEXT-"))
+    (let* ((n-context (gensym "N-CONTEXT-"))
            (context-temp (make-lambda-var :%source-name n-context
                                           :arg-info (make-arg-info :kind :more-context)))
-           (n-count (sb-xc:gensym "N-COUNT-"))
+           (n-count (gensym "N-COUNT-"))
            (count-temp (make-lambda-var :%source-name n-count
                                         :type (specifier-type 'index)
                                         :arg-info (make-arg-info :kind :more-count))))
@@ -470,11 +479,11 @@
       ;; and take advantage of the base+index+displacement addressing
       ;; mode on x86oids.)
       (when (optional-dispatch-keyp res)
-        (let ((n-index (sb-xc:gensym "N-INDEX-"))
-              (n-key (sb-xc:gensym "N-KEY-"))
-              (n-value-temp (sb-xc:gensym "N-VALUE-TEMP-"))
-              (n-allowp (sb-xc:gensym "N-ALLOWP-"))
-              (n-lose (sb-xc:gensym "N-LOSE-"))
+        (let ((n-index (gensym "N-INDEX-"))
+              (n-key (gensym "N-KEY-"))
+              (n-value-temp (gensym "N-VALUE-TEMP-"))
+              (n-allowp (gensym "N-ALLOWP-"))
+              (n-lose (gensym "N-LOSE-"))
               (allowp (or (optional-dispatch-allowp res)
                           (policy *lexenv* (zerop safety))))
               (found-allow-p nil))
@@ -498,9 +507,9 @@
                      (keyword (arg-info-key info))
                      (supplied-p (arg-info-supplied-p info))
                      (supplied-used-p (arg-info-supplied-used-p info))
-                     (n-value (sb-xc:gensym "N-VALUE-"))
+                     (n-value (gensym "N-VALUE-"))
                      (clause (cond (supplied-p
-                                    (let ((n-supplied (sb-xc:gensym "N-SUPPLIED-")))
+                                    (let ((n-supplied (gensym "N-SUPPLIED-")))
                                       (temps (list n-supplied
                                                    (if supplied-used-p
                                                        nil
@@ -572,7 +581,7 @@
                     (case ,n-key ,@(tests))))))
 
             (unless allowp
-              (let ((location (opaquely-quote (make-restart-location))))
+              (let ((location (make-restart-location)))
                 (body `(if (and (not (unbound-marker-p ,n-lose))
                                 (not ,n-allowp))
                            (%unknown-key-arg-error ,n-lose ,location)
@@ -627,10 +636,10 @@
         ;; Make up two extra variables, and squirrel them away in
         ;; ARG-INFO-DEFAULT for transforming (VALUES-LIST REST) into
         ;; (%MORE-ARG-VALUES CONTEXT 0 COUNT) when possible.
-        (let* ((context-name (sb-xc:gensym "REST-CONTEXT-"))
+        (let* ((context-name (gensym "REST-CONTEXT-"))
                (context (make-lambda-var :%source-name context-name
                                          :arg-info (make-arg-info :kind :more-context)))
-               (count-name (sb-xc:gensym "REST-COUNT-"))
+               (count-name (gensym "REST-COUNT-"))
                (count (make-lambda-var :%source-name count-name
                                        :arg-info (make-arg-info :kind :more-count)
                                        :type (specifier-type 'index))))
@@ -658,7 +667,7 @@
         (main-vars val-temp)
         (bind-vars key)
         (cond ((or hairy-default supplied-p)
-               (let* ((n-supplied (sb-xc:gensym "N-SUPPLIED-"))
+               (let* ((n-supplied (gensym "N-SUPPLIED-"))
                       (supplied-temp (make-lambda-var
                                       :%source-name n-supplied)))
                  (unless supplied-p
@@ -869,10 +878,21 @@
 
     res))
 
+(defvar *lambda-conversions*)
+
+(defun add-types-for-fixed-args (fun vars)
+  (let ((fun-info (info :function :info fun)))
+    (when (and fun-info
+               (ir1-attributep (fun-info-attributes fun-info) fixed-args))
+      (loop for type in (fun-type-required (info :function :type fun))
+            for var in vars
+            do (setf (lambda-var-type var) type))))
+  vars)
+
 ;;; Convert a LAMBDA form into a LAMBDA leaf or an OPTIONAL-DISPATCH leaf.
 (defun ir1-convert-lambda (form &key (source-name '.anonymous.)
-                           debug-name maybe-add-debug-catch
-                           system-lambda)
+                                     debug-name maybe-add-debug-catch
+                                     system-lambda)
   (unless (consp form)
     (compiler-error "A ~S was found when expecting a lambda expression:~%  ~S"
                     (type-of form)
@@ -890,56 +910,62 @@
     (bug "Both SYSTEM-LAMBDA and MAYBE-ADD-DEBUG-CATCH specified"))
   (unless (or debug-name (neq '.anonymous. source-name))
     (setf debug-name (name-lambdalike form)))
-  (multiple-value-bind (forms decls doc) (parse-body (cddr form) t)
-    (let ((*lexenv* (process-muffle-decls decls *lexenv*)))
-      (multiple-value-bind (vars keyp allow-other-keys aux-vars aux-vals)
-          (make-lambda-vars (cadr form))
-        (binding* (((*lexenv* result-type post-binding-lexenv
-                              lambda-list explicit-check source-form)
-                    (process-decls decls (append aux-vars vars) nil
-                                   :binding-form-p t :allow-lambda-list t))
-                   (debug-catch-p (and maybe-add-debug-catch
-                                       *allow-instrumenting*
-                                       (policy *lexenv*
-                                           (>= insert-debug-catch 2))))
-                   (forms (if debug-catch-p
-                              (wrap-forms-in-debug-catch forms)
-                              forms))
-                   (forms (if (eq result-type *wild-type*)
-                              forms
-                              `((the ,(type-specifier result-type) (progn ,@forms)))))
-                   (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
-                   (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
-                               (ir1-convert-hairy-lambda forms vars keyp
-                                                         allow-other-keys
-                                                         aux-vars aux-vals
-                                                         :post-binding-lexenv post-binding-lexenv
-                                                         :source-name source-name
-                                                         :debug-name debug-name
-                                                         :system-lambda system-lambda))
-                              (t
-                               (ir1-convert-lambda-body forms vars
-                                                        :aux-vars aux-vars
-                                                        :aux-vals aux-vals
-                                                        :post-binding-lexenv post-binding-lexenv
-                                                        :source-name source-name
-                                                        :debug-name debug-name
-                                                        :system-lambda system-lambda)))))
-          (when explicit-check
-            (setf (getf (functional-plist res) 'explicit-check) explicit-check))
-          (setf (functional-inline-expansion res) (or source-form form))
-          (setf (functional-arg-documentation res)
-                (if (eq lambda-list :unspecified)
-                    (strip-lambda-list (cadr form) :arglist)
-                    lambda-list))
-          (setf (functional-documentation res) doc)
-          (when (boundp '*lambda-conversions*)
-            ;; KLUDGE: Not counting TL-XEPs is a lie, of course, but
-            ;; keeps things less confusing to users of TIME, where this
-            ;; count gets used.
-            (unless (and (consp debug-name) (eq 'tl-xep (car debug-name)))
-              (incf *lambda-conversions*)))
-          res)))))
+  (binding* (((forms decls doc) (parse-body (cddr form) t))
+             ((*lexenv* source-form) (process-muffle-decls decls *lexenv*))
+             (*current-path* (or (and source-form
+                                      (get-source-path source-form))
+                                 *current-path*))
+             ((vars keyp allow-other-keys aux-vars aux-vals)
+              (make-lambda-vars (cadr form)))
+             ((*lexenv* result-type post-binding-lexenv
+                        lambda-list explicit-check source-form
+                        local-policy)
+              (process-decls decls (append aux-vars vars) nil
+                             :binding-form-p t :allow-lambda-list t))
+             (debug-catch-p (and maybe-add-debug-catch
+                                 *allow-instrumenting*
+                                 (policy *lexenv*
+                                     (>= insert-debug-catch 2))))
+             (forms (if debug-catch-p
+                        (wrap-forms-in-debug-catch forms)
+                        forms))
+             (forms (if (eq result-type *wild-type*)
+                        forms
+                        `((the ,(type-specifier result-type) (progn ,@forms)))))
+             (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
+             (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
+                         (ir1-convert-hairy-lambda forms vars keyp
+                                                   allow-other-keys
+                                                   aux-vars aux-vals
+                                                   :post-binding-lexenv post-binding-lexenv
+                                                   :source-name source-name
+                                                   :debug-name debug-name
+                                                   :system-lambda system-lambda))
+                        (t
+                         (ir1-convert-lambda-body forms
+                                                  (add-types-for-fixed-args source-name vars)
+                                                  :aux-vars aux-vars
+                                                  :aux-vals aux-vals
+                                                  :post-binding-lexenv post-binding-lexenv
+                                                  :source-name source-name
+                                                  :debug-name debug-name
+                                                  :system-lambda system-lambda
+                                                  :local-policy local-policy)))))
+    (when explicit-check
+      (setf (getf (functional-plist res) 'explicit-check) explicit-check))
+    (setf (functional-inline-expansion res) (or source-form form))
+    (setf (functional-arg-documentation res)
+          (if (eq lambda-list :unspecified)
+              (strip-lambda-list (cadr form) :arglist)
+              lambda-list))
+    (setf (functional-documentation res) doc)
+    (when (boundp '*lambda-conversions*)
+      ;; KLUDGE: Not counting TL-XEPs is a lie, of course, but
+      ;; keeps things less confusing to users of TIME, where this
+      ;; count gets used.
+      (unless (and (consp debug-name) (eq 'tl-xep (car debug-name)))
+        (incf *lambda-conversions*)))
+    res))
 
 (defun wrap-forms-in-debug-catch (forms)
   #+unwind-to-frame-and-call-vop
@@ -997,7 +1023,16 @@
     ((named-lambda)
      (let* ((name (cadr thing))
             (lambda-expression `(lambda ,@(cddr thing)))
-            (*inline-expansions* (list name 1 *inline-expansions*)))
+            (*inline-expansions* (list name 1 *inline-expansions*))
+            (simple-lexenv-p (simple-lexical-environment-p *lexenv*)))
+       ;; Discard any forward references to this function unless we
+       ;; are block compiling and the lexical environment doesn't
+       ;; contain any hair. If the lexical environment is too hairy, then we only
+       ;; install the definition during the processing of this NAMED-LAMBDA,
+       ;; ensuring that the function cannot be called outside of the correct
+       ;; environment.
+       (unless (and (block-compile *compilation*) simple-lexenv-p)
+         (remhash name (free-funs *ir1-namespace*)))
        (if (and name (legal-fun-name-p name))
            (let ((defined-fun-res (get-defined-fun name (second lambda-expression)))
                  (res (ir1-convert-lambda lambda-expression
@@ -1005,13 +1040,19 @@
                                           :source-name name))
                  (info (info :function :info name)))
              (setf (functional-inlinep res) (info :function :inlinep name)
-                   (defined-fun-named-lambda-p defined-fun-res) t)
+                   (defined-fun-same-block-p defined-fun-res) t)
              (when (has-toplevelness-decl lambda-expression)
                (setf (functional-top-level-defun-p res) t))
              ;; FIXME: Should non-entry block compiled defuns have
              ;; this propagate?
-             (assert-global-function-definition-type name res)
-             (push res (defined-fun-functionals defined-fun-res))
+             (unless (and info
+                          (ir1-attributep (fun-info-attributes info) fixed-args))
+               (assert-global-function-definition-type name res))
+             ;; If in a simple environment, then we can allow
+             ;; backward references to this function from following
+             ;; top-level forms.
+             (when simple-lexenv-p
+               (push res (defined-fun-functionals defined-fun-res)))
              (unless (or
                       (eq (defined-fun-inlinep defined-fun-res) 'notinline)
                       ;; Don't treat recursive stubs like CAR as self-calls
@@ -1043,8 +1084,8 @@
 
 (declaim (end-block))
 
-
 ;;;; defining global functions
+
 ;;; Given a lambda-list, return a FUN-TYPE object representing the signature:
 ;;; return type is *, and each individual arguments type is T -- but we get
 ;;; the argument counts and keywords.
@@ -1066,112 +1107,59 @@
             (allow (when (ll-kwds-allowp llks) '(&allow-other-keys))))
         (careful-specifier-type `(function (,@reqs ,@opts ,@rest ,@keys ,@allow) *))))))
 
-#-sb-devel
-(declaim (start-block maybe-inline-syntactic-closure))
-
-;;; Take the lexenv surrounding an inlined function and extract things
-;;; needed for the inline expansion suitable for dumping into fasls.
-;;; Right now it's MACROLET, SYMBOL-MACROLET, SPECIAL and
-;;; INLINE/NOTINLINE declarations. Upon encountering something else return NIL.
-;;; This is later used by PROCESS-INLINE-LEXENV to reproduce the lexenv.
-;;;
-;;; Previously it just used the functions and vars of the innermost
-;;; lexenv, but the body of macrolet can refer to other macrolets
-;;; defined earlier, so it needs to process all the parent lexenvs to
-;;; recover the proper order.
-(defun reconstruct-lexenv (lexenv)
-  (let (shadowed-funs
-        shadowed-vars
-        result)
-    (loop for env = lexenv then parent
-          for parent = (lexenv-parent env)
-          for vars = (lexenv-vars env)
-          for funs = (lexenv-funs env)
-          for declarations = nil
-          for symbol-macros = nil
-          for macros = nil
-          do
-          (loop for binding in vars
-                for (name . what) = binding
-                unless (and parent
-                            (find binding (lexenv-vars parent)))
-                do (typecase what
-                     (cons
-                      (aver (eq (car what) 'macro))
-                      (push name shadowed-vars)
-                      (push (list name (cdr what)) symbol-macros))
-                     (global-var
-                      (aver (eq (global-var-kind what) :special))
-                      (push `(special ,name) declarations))
-                     (t
-                      (unless (memq name shadowed-vars)
-                        (return-from reconstruct-lexenv)))))
-          (loop for binding in funs
-                for (name . what) = binding
-                unless (and parent
-                            (find binding (lexenv-funs parent)))
-                do
-                (typecase what
-                  (cons
-                   (push name shadowed-funs)
-                   (let ((expression (function-lambda-expression (cdr what))))
-                     (aver expression)
-                     (push (cons name expression) macros)))
-                  ;; FIXME: Is there a good reason for this not to be
-                  ;; DEFINED-FUN (which :INCLUDEs GLOBAL-VAR, in case
-                  ;; you're wondering how this ever worked :-)? Maybe
-                  ;; in conjunction with an AVERrance that it's not an
-                  ;; (AND GLOBAL-VAR (NOT GLOBAL-FUN))? -- CSR,
-                  ;; 2002-07-08
-                  (global-var
-                   (unless (defined-fun-p what)
-                     (return-from reconstruct-lexenv))
-                   (push `(,(car (defined-fun-inlinep what))
-                           ,name)
-                         declarations))
-                  (t
-                   (unless (memq name shadowed-funs)
-                     (return-from reconstruct-lexenv)))))
-          (when declarations
-            (setf result (list* :declare declarations (and result (list result)))))
-          (when symbol-macros
-            (setf result (list* :symbol-macro symbol-macros (and result (list result)))))
-          (when macros
-            (setf result (list* :macro macros (and result (list result)))))
-          while (and parent
-                     (not (null-lexenv-p parent))))
-    result))
-
-;;; Return a sexpr for LAMBDA in LEXENV such that loading it from fasl
-;;; preserves the original lexical environment for inlining.
-;;; Return NIL if the lexical environment is too complicated.
-(defun maybe-inline-syntactic-closure (lambda lexenv)
+;;; Return a lambda form that has been "closed" with respect to
+;;; LEXENV, returning a LAMBDA-WITH-LEXENV if there are interesting
+;;; declarations. To handle local macros, rather than closing over
+;;; definitions in the environment, expand all macros in the body of
+;;; LAMBDA, so that nothing in the syntactic environment is needed in
+;;; the expansion. If there is something too complex in the lexical
+;;; environment (like a lexical variable), then we return NIL.
+(defun inline-syntactic-closure-lambda (lambda lexenv)
   (declare (type list lambda) (type lexenv-designator lexenv))
   (aver (eql (first lambda) 'lambda))
-  ;; We used to have a trivial implementation, verifying that lexenv
-  ;; was effectively null. However, this fails to take account of the
-  ;; idiom
-  ;;
-  ;; (declaim (inline foo))
-  ;; (macrolet ((def (x) `(defun ,x () ...)))
-  ;;   (def foo))
-  ;;
-  ;; which, while too complicated for the cross-compiler to handle in
-  ;; unfriendly foreign lisp environments, would be good to support in
-  ;; the target compiler. -- CSR, 2002-05-13 and 2002-11-02
   (typecase lexenv
    (lexenv
     (let ((vars (lexenv-vars lexenv))
-          (funs (lexenv-funs lexenv)))
-      (acond ((or (lexenv-blocks lexenv) (lexenv-tags lexenv)) nil)
-             ((and (null vars) (null funs)) lambda)
-             ;; If the lexenv is too hairy for cross-compilation,
-             ;; you'll find out later, when trying to perform inlining.
-             ;; This is fine, because if the inline expansion is only
-             ;; for the target, it's totally OK to cross-compile this
-             ;; defining form. The syntactic env is correctly captured.
-             ((reconstruct-lexenv lexenv)
-              `(lambda-with-lexenv ,it ,@(cdr lambda))))))
+          (funs (lexenv-funs lexenv))
+          (decls ()))
+      (cond ((or (lexenv-blocks lexenv) (lexenv-tags lexenv)) nil)
+            ((and (null vars) (null funs)) lambda)
+            ((dolist (x vars nil)
+               (let ((name (car x))
+                     (what (cdr x)))
+                 (when (eq x (assoc name vars :test #'eq))
+                   (typecase what
+                     (cons
+                      (aver (eq (car what) 'macro)))
+                     (global-var
+                      (aver (eq (global-var-kind what) :special))
+                      (push `(special ,name) decls))
+                     (t (return t))))))
+             nil)
+            ((dolist (x funs nil)
+               (let ((name (car x))
+                     (what (cdr x)))
+                 (when (eq x (assoc name funs :test #'equal))
+                   (typecase what
+                     (cons)
+                     ;; FIXME: Is there a good reason for this not to be
+                     ;; DEFINED-FUN (which :INCLUDEs GLOBAL-VAR, in case
+                     ;; you're wondering how this ever worked :-)? Maybe
+                     ;; in conjunction with an AVERrance that it's not an
+                     ;; (AND GLOBAL-VAR (NOT GLOBAL-FUN))? -- CSR,
+                     ;; 2002-07-08
+                     (global-var
+                      (when (defined-fun-p what)
+                        (push `(,(car (defined-fun-inlinep what))
+                                ,name)
+                              decls)))
+                     (t (return t))))))
+             nil)
+            (t
+             (let ((expansion (sb-walker:macroexpand-all lambda lexenv)))
+               (if decls
+                   `(lambda-with-lexenv (:declare ,decls) ,@(cdr expansion))
+                   expansion))))))
    #+(and sb-fasteval (not sb-xc-host))
    (sb-interpreter:basic-env
     (awhen (sb-interpreter::reconstruct-syntactic-closure-env lexenv)
@@ -1179,9 +1167,6 @@
    #+sb-fasteval
    (null lambda))) ; trivial case. Never occurs in the compiler.
 
-(declaim (end-block))
-
-#-sb-devel
 (declaim (start-block ir1-convert-inline-lambda))
 
 ;;; Convert the forms produced by RECONSTRUCT-LEXENV to LEXENV
@@ -1250,7 +1235,7 @@
                 (lexenv-flushable *lexenv*)
                 lexenv-lambda
                 *lexenv*)))
-         (*inlining* (not *transforming*))
+         (*inlining* (1+ *inlining*))
          (clambda (ir1-convert-lambda body
                                       :source-name source-name
                                       :debug-name debug-name
@@ -1331,6 +1316,18 @@
                 "previous declaration"
                 "previous definition"))))
 
+;;; The lexical environment is hairy if it has stuff like lexical
+;;; variables, blocks, or tags that are not load-time constants. Local
+;;; functions and macros are OK, since we are worried about outside
+;;; calls. Macros get expanded, and local functions are load-time
+;;; constants.
+(defun simple-lexical-environment-p (lexenv)
+  (and (null (lexenv-blocks lexenv))
+       (null (lexenv-tags lexenv))
+       (every (lambda (entry)
+                (consp (cdr entry)))
+              (lexenv-vars lexenv))))
+
 ;;; Used for global inline expansion. Earlier something like this was
 ;;; used by %DEFUN too. FIXME: And now it's probably worth rethinking
 ;;; whether this function is a good idea at all.
@@ -1370,16 +1367,17 @@
         (note-inlining (optional-dispatch-more-entry fun))
         (mapc #'note-inlining (optional-dispatch-entry-points fun))))
     ;; substitute for any old references
-    (unless (or (eq (defined-fun-inlinep var) :notinline)
+    (unless (or (eq (defined-fun-inlinep var) 'notinline)
                 (not (block-compile *compilation*))
                 (and info
                      (or (fun-info-transforms info)
                          (fun-info-templates info)
                          (fun-info-ir2-convert info))))
       (substitute-leaf fun var)
-      ;; If in a simple environment, then we can allow backward references
-      ;; to this function from following top-level forms.
-      (when expansion
+      ;; If in a simple environment, then we can allow backward
+      ;; references to this function from following top-level
+      ;; forms.
+      (when (simple-lexical-environment-p *lexenv*)
         (push fun (defined-fun-functionals var))))
     fun))
 
@@ -1467,25 +1465,29 @@ is potentially harmful to any already-compiled callers using (SAFETY 0)."
 ;;; since those can always be reconstructed from a defstruct description.
 (defun %compiler-defun (name compile-toplevel inline-lambda extra-info)
   (let ((defined-fun nil)) ; will be set below if we're in the compiler
-    (when compile-toplevel
-      (with-single-package-locked-error
-          (:symbol name "defining ~S as a function")
-        (setf defined-fun
-              ;; Try to pass the lambda-list to GET-DEFINED-FUN if we can.
-              (if (atom inline-lambda)
-                  (get-defined-fun name)
-                  (get-defined-fun
-                   name (ecase (car inline-lambda)
-                         (lambda-with-lexenv (third inline-lambda))
-                         (lambda (second inline-lambda)))))))
-      (when (boundp '*lexenv*)
-        (aver (producing-fasl-file))
-        (if (member name (fun-names-in-this-file *compilation*) :test #'equal)
-            (warn 'duplicate-definition :name name)
-            (push name (fun-names-in-this-file *compilation*))))
-      ;; I don't know why this is guarded by (WHEN compile-toplevel),
-      ;; because regular old %DEFUN is going to call this anyway.
-      (%set-inline-expansion name defined-fun inline-lambda extra-info))
+    (cond (compile-toplevel
+           (with-single-package-locked-error
+               (:symbol name "defining ~S as a function")
+             (setf defined-fun
+                   ;; Try to pass the lambda-list to GET-DEFINED-FUN if we can.
+                   (if (atom inline-lambda)
+                       (get-defined-fun name)
+                       (get-defined-fun
+                        name (ecase (car inline-lambda)
+                               (lambda-with-lexenv (third inline-lambda))
+                               (lambda (second inline-lambda)))))))
+           (when (boundp '*lexenv*)
+             (aver (producing-fasl-file))
+             (if (member name (fun-names-in-this-file *compilation*) :test #'equal)
+                 (warn 'duplicate-definition :name name)
+                 (push name (fun-names-in-this-file *compilation*))))
+           ;; I don't know why this is guarded by (WHEN compile-toplevel),
+           ;; because regular old %DEFUN is going to call this anyway.
+           (%set-inline-expansion name defined-fun inline-lambda extra-info))
+          ((boundp 'sb-fasl::*current-fasl-group*)
+           (if (member name (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*) :test #'equal)
+               (warn 'duplicate-definition :name name)
+               (push name (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*)))))
 
     (become-defined-fun-name name)
 
