@@ -667,7 +667,7 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
 
 (declaim (inline make-callable))
 (defun make-callable (generator method-alist wrappers)
-  (function-funcall generator method-alist wrappers))
+  (funcall (the function generator) method-alist wrappers))
 
 (defun make-dispatch-dfun (gf)
   (values (get-dispatch-function gf) nil (dispatch-dfun-info)))
@@ -715,16 +715,20 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
     (when (and ,applicable (not (memq ,gf *dfun-miss-gfs-on-stack*)))
       (let ((*dfun-miss-gfs-on-stack* (cons ,gf *dfun-miss-gfs-on-stack*)))
         ,@body))
-    ;; Create a FAST-INSTANCE-BOUNDP structure instance for a cached
-    ;; SLOT-BOUNDP so that INVOKE-EMF does the right thing, that is,
-    ;; does not signal a SLOT-UNBOUND error for a boundp test.
-    ,@(if type
-          ;; FIXME: could the NEMF not be a CONS (for :CLASS-allocated
-          ;; slots?)
-          `((if (and (eq ,type 'boundp) (integerp ,nemf))
-                (invoke-emf (make-fast-instance-boundp :index ,nemf) ,args)
-                (invoke-emf ,nemf ,args)))
-          `((invoke-emf ,nemf ,args)))))
+    ,(if type
+         ;; Munge the EMF so that INVOKE-EMF can do the right thing:
+         ;; BOUNDP gets a structure, WRITER the logical not of the
+         ;; index, so that READER can use the raw index.
+         ;;
+         ;; FIXME: could the NEMF not be a CONS (for :CLASS-allocated
+         ;; slots?)
+         `(if (integerp ,nemf)
+              (case ,type
+                (boundp (invoke-emf (make-fast-instance-boundp :index ,nemf) ,args))
+                (reader (invoke-emf ,nemf ,args))
+                (writer (invoke-emf (lognot ,nemf) ,args)))
+              (invoke-emf ,nemf ,args))
+         `(invoke-emf ,nemf ,args))))
 
 ;;; The dynamically adaptive method lookup algorithm is implemented is
 ;;; implemented as a kind of state machine. The kinds of
@@ -812,10 +816,8 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
           ((every (lambda (method)
                     (if (consp method)
                         (let ((class (early-method-class method)))
-                          (or (eq class *the-class-standard-boundp-method*)
-                              (eq class *the-class-global-boundp-method*)))
-                        (or (standard-boundp-method-p method)
-                            (global-boundp-method-p method))))
+                          (eq class *the-class-global-boundp-method*))
+                        (global-boundp-method-p method)))
                   methods)
            'boundp)
           ((every (lambda (method)
@@ -885,8 +887,6 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
           (t
            (make-final-caching-dfun gf classes-list new-class)))))
 
-(define-load-time-global *pcl-misc-random-state* (make-random-state))
-
 (defun accessor-miss (gf new object dfun-info)
   (let* ((ostate (type-of dfun-info))
          (otype (dfun-info-accessor-type dfun-info))
@@ -900,7 +900,7 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
       ;; which are the parameters of the new state, and get other
       ;; information from the lexical variables bound above.
       (flet ((two-class (index w0 w1)
-               (when (zerop (random 2 *pcl-misc-random-state*))
+               (when (zerop (random 2 (load-time-value *pcl-misc-random-state*)))
                  (psetf w0 w1 w1 w0))
                (dfun-update gf
                             #'make-two-class-accessor-dfun
@@ -1697,6 +1697,15 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
 (defun update-dfun (generic-function &optional dfun cache info)
   (let ((early-p (early-gf-p generic-function)))
     (flet ((update ()
+             ;; If GENERIC-FUNCTION has a CALL-NEXT-METHOD argument
+             ;; checker, the methods of the checker (the checker is a
+             ;; generic function, each method caches a computation for
+             ;; a combination of original and C-N-M argument classes)
+             ;; must be re-computed.
+             (when (eq **boot-state** 'complete)
+               (let ((checker (gf-info-cnm-checker (gf-arg-info generic-function))))
+                 (when checker
+                   (remove-methods checker))))
              ;; Save DFUN-STATE, so that COMPUTE-DISCRIMINATING-FUNCTION can
              ;; access it, and so that it's there for eg. future cache updates.
              (set-dfun generic-function dfun cache info)
