@@ -34,21 +34,23 @@
 
 (defun dfun-arg-symbol (arg-number)
   (or (nth arg-number *dfun-arg-symbols*)
-      (format-symbol *pcl-package* ".ARG~A." arg-number)))
+      (pcl-format-symbol ".ARG~A." arg-number)))
 
 (declaim (list *slot-vector-symbols*))
 (define-load-time-global *slot-vector-symbols* '(.SLOTS0. .SLOTS1. .SLOTS2. .SLOTS3.))
 
 (defun slot-vector-symbol (arg-number)
   (or (nth arg-number *slot-vector-symbols*)
-      (format-symbol *pcl-package* ".SLOTS~A." arg-number)))
+      (pcl-format-symbol ".SLOTS~A." arg-number)))
 
 (declaim (inline make-dfun-required-args))
 (defun make-dfun-required-args (count)
   (declare (type index count))
-  (let (result)
-    (dotimes (i count (nreverse result))
-      (push (dfun-arg-symbol i) result))))
+  ;; N.B.: don't PUSH and NREVERSE here. COLLECT will cons in the system TLAB,
+  ;; but NREVERSE won't because we don't inline NREVERSE.
+  (collect ((result))
+    (dotimes (i count (result))
+      (result (dfun-arg-symbol i)))))
 
 (defun make-dfun-lambda-list (nargs applyp)
   (let ((required (make-dfun-required-args nargs)))
@@ -71,7 +73,7 @@
     ;; of the vop operand restrictions or something that I don't understand.
     ;; Which is to say, PCL compilation reliably broke when changed to INDEX.
     (if applyp
-        (values (append required '(&more .more-context. .more-count.))
+        (values (sb-impl::sys-tlab-append required '(&more .more-context. .more-count.))
                 required
                 '((sb-c:%listify-rest-args
                    .more-context. (the (and unsigned-byte fixnum)
@@ -182,12 +184,12 @@
                    (declare (optimize (sb-c:store-source-form 0)))
                    (declare (optimize (sb-c::store-closure-debug-pointer 3)))
                    #'(lambda ,args
-                       (let ()
+                       (let () ; What is this LET doing?
                          (declare #.*optimize-speed*)
                          ,form)))))
     (values (if *precompiling-lap*
                 `#',lambda
-                (compile nil lambda))
+                (pcl-compile lambda :safe))
             nil)))
 
 ;;; note on implementation for CMU 17 and later (including SBCL):
@@ -219,15 +221,15 @@
                                ,@(unless class-slot-p
                                    `((setq slots
                                            (std-instance-slots ,instance))))
-                               (%instance-layout ,instance))
+                               (%instance-wrapper ,instance))
                               ((fsc-instance-p ,instance)
                                ,@(unless class-slot-p
                                    `((setq slots
                                            (fsc-instance-slots ,instance))))
-                               (%funcallable-instance-layout ,instance)))))
+                               (%fun-wrapper ,instance)))))
         (block access
           (when (and wrapper
-                     (not (zerop (layout-clos-hash wrapper)))
+                     (not (zerop (wrapper-clos-hash wrapper)))
                      ,@(if (eql 1 1-or-2-class)
                            `((eq wrapper wrapper-0))
                            `((or (eq wrapper wrapper-0)
@@ -304,7 +306,7 @@
 (defun emit-miss (miss-fn args applyp)
   (if applyp
       `(multiple-value-call ,miss-fn ,@args
-                            (sb-c::%more-arg-values .more-context.
+                            (sb-c:%more-arg-values .more-context.
                                                     0
                                                     .more-count.))
       `(funcall ,miss-fn ,@args)))
@@ -350,9 +352,7 @@
          (wrapper-bindings (mapcan (lambda (arg mt)
                                      (unless (eq mt t)
                                        (incf index)
-                                       `((,(format-symbol *pcl-package*
-                                                          "WRAPPER-~D"
-                                                          index)
+                                       `((,(pcl-format-symbol "WRAPPER-~D" index)
                                           ,(emit-fetch-wrapper
                                             mt arg miss-tag (pop slot-vars))))))
                                    args metatypes))
@@ -385,18 +385,18 @@
      (with-unique-names (wrapper)
        `(cond ((std-instance-p ,argument)
                ,(if slots-var
-                    `(let ((,wrapper (%instance-layout ,argument)))
-                       (when (layout-for-std-class-p ,wrapper)
+                    `(let ((,wrapper (%instance-wrapper ,argument)))
+                       (when (layout-for-pcl-obj-p ,wrapper)
                          (setq ,slots-var (std-instance-slots ,argument)))
                        ,wrapper)
-                    `(%instance-layout ,argument)))
+                    `(%instance-wrapper ,argument)))
               ((fsc-instance-p ,argument)
                ,(if slots-var
-                    `(let ((,wrapper (%funcallable-instance-layout ,argument)))
-                       (when (layout-for-std-class-p ,wrapper)
+                    `(let ((,wrapper (%fun-wrapper ,argument)))
+                       (when (layout-for-pcl-obj-p ,wrapper)
                          (setq ,slots-var (fsc-instance-slots ,argument)))
                        ,wrapper)
-                    `(%funcallable-instance-layout ,argument)))
+                    `(%fun-wrapper ,argument)))
                (t (go ,miss-tag)))))
     ;; Sep92 PCL used to distinguish between some of these cases (and
     ;; spuriously exclude others).  Since in SBCL
@@ -407,7 +407,7 @@
      (when slots-var
        (bug "SLOT requested for metatype ~S, but it isn't going to happen."
             metatype))
-     `(layout-of ,argument))
+     `(wrapper-of ,argument))
     ;; a metatype of NIL should never be seen here, as NIL is only in
     ;; the metatypes before a generic function is fully initialized.
     ;; T should never be seen because we never need to get a wrapper
