@@ -21,24 +21,32 @@
 #define ALIEN_STACK_SIZE (1024*1024)     /* chosen at random */
 
 /* eventually choosable per-thread: */
+#ifndef DEFAULT_CONTROL_STACK_SIZE // CFLAGS=-D... can set this
 #define DEFAULT_CONTROL_STACK_SIZE (2*1024*1024)
+#endif
 
 /* constants derived from the fundamental constants in passed by GENESIS */
-#ifdef LISP_FEATURE_CHENEYGC
-#define DEFAULT_DYNAMIC_SPACE_SIZE (DYNAMIC_0_SPACE_END - DYNAMIC_0_SPACE_START)
-#endif
 #define READ_ONLY_SPACE_SIZE (READ_ONLY_SPACE_END - READ_ONLY_SPACE_START)
 #define STATIC_SPACE_SIZE (STATIC_SPACE_END - STATIC_SPACE_START)
-#define IMMOBILE_SPACE_SIZE (IMMOBILE_SPACE_END - IMMOBILE_SPACE_START)
-#ifdef LISP_FEATURE_LINKAGE_TABLE
-#define LINKAGE_TABLE_SPACE_SIZE \
-    (LINKAGE_TABLE_SPACE_END - LINKAGE_TABLE_SPACE_START)
+
+#ifdef LISP_FEATURE_DARWIN_JIT
+#define STATIC_CODE_SPACE_SIZE (STATIC_CODE_SPACE_END - STATIC_CODE_SPACE_START)
 #endif
 
-#if !defined(LANGUAGE_ASSEMBLY)
+#define ALIEN_LINKAGE_TABLE_SPACE_END \
+    (ALIEN_LINKAGE_TABLE_SPACE_START + ALIEN_LINKAGE_TABLE_SPACE_SIZE)
+
+#if !defined(__ASSEMBLER__)
 #include "thread.h"
 
-#ifdef LISP_FEATURE_STACK_GROWS_DOWNWARD_NOT_UPWARD
+#if defined(LISP_FEATURE_WIN32)
+
+#define CONTROL_STACK_HARD_GUARD_PAGE(th) \
+    ((os_vm_address_t)(th->control_stack_start))
+#define CONTROL_STACK_GUARD_PAGE(th) \
+    (CONTROL_STACK_HARD_GUARD_PAGE(th) + win32_page_size + win32_stack_guarantee)
+
+#elif defined(LISP_FEATURE_STACK_GROWS_DOWNWARD_NOT_UPWARD)
 
 #define CONTROL_STACK_HARD_GUARD_PAGE(th) \
     ((os_vm_address_t)(th->control_stack_start))
@@ -46,6 +54,7 @@
     (CONTROL_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size)
 #define CONTROL_STACK_RETURN_GUARD_PAGE(th) \
     (CONTROL_STACK_GUARD_PAGE(th) + os_vm_page_size)
+
 #else
 
 #define CONTROL_STACK_HARD_GUARD_PAGE(th) \
@@ -57,7 +66,7 @@
 
 #endif
 
-#ifdef ALIEN_STACK_GROWS_DOWNWARD
+#ifndef ALIEN_STACK_GROWS_UPWARD
 
 #define ALIEN_STACK_HARD_GUARD_PAGE(th)         \
     ((os_vm_address_t)(th->alien_stack_start))
@@ -66,7 +75,7 @@
 #define ALIEN_STACK_RETURN_GUARD_PAGE(th) \
     (ALIEN_STACK_GUARD_PAGE(th) + os_vm_page_size)
 
-#elif defined(ALIEN_STACK_GROWS_UPWARD)
+#else
 
 #define ALIEN_STACK_HARD_GUARD_PAGE(th)                            \
     (((os_vm_address_t)th->alien_stack_start) + ALIEN_STACK_SIZE - \
@@ -76,8 +85,6 @@
 #define ALIEN_STACK_RETURN_GUARD_PAGE(th) \
     (ALIEN_STACK_GUARD_PAGE(th) - os_vm_page_size)
 
-#else
-#error ALIEN_STACK_GROWS_DOWNWARD or ALIEN_STACK_GROWS_UPWARD has to be defined
 #endif
 
 #define BINDING_STACK_HARD_GUARD_PAGE(th)                              \
@@ -88,7 +95,8 @@
 #define BINDING_STACK_RETURN_GUARD_PAGE(th) \
     (BINDING_STACK_GUARD_PAGE(th) - os_vm_page_size)
 
-extern void allocate_spaces(void);
+extern void allocate_lisp_dynamic_space(boolean);
+extern boolean allocate_hardwired_spaces(boolean);
 
 extern void
 protect_control_stack_hard_guard_page(int protect_p, struct thread *thread);
@@ -120,7 +128,7 @@ extern os_vm_address_t undefined_alien_address;
  *   #include "x86-validate.h"
  *   #endif
  * and so forth. In SBCL, the memory map data are defined at the Lisp
- * level (compiler/target/parms.lisp) and stuffed into the sbcl.h file
+ * level (compiler/{arch}/parms.lisp) and stuffed into the sbcl.h file
  * created by GENESIS, so there's no longer a need for an
  * architecture-dependent header file of memory map data.
  */
